/* file KUT/kut.h */
/* Time-stamp: <1998 Augu 30 Sun 11:16:51 CEST Basile.Starynkevitch@wanadoo.fr> */
/* rcsid: $Id: kut.h,v 0.8 1998/08/30 09:19:02 basile Exp $ */

#ifndef __KUT_INCLUDED__
#define  __KUT_INCLUDED__ 1

#define KUT_H_IDENT "@(#)$Id: kut.h,v 0.8 1998/08/30 09:19:02 basile Exp $"

/* we optimize the usual case when the Port is a compiletime constant */
/* inportb() =	Get a byte from an I/O port */
#define inportb(Port) 						\
({ register unsigned char res;					\
  if (__builtin_constant_p(Port)) 				\
    asm volatile(" inb %1, %%al\n" : "=a" (res) : "i" (Port));	\
  else								\
    asm volatile(" inb %%dx,%%al\n" : "=a" (res) : "d" (Port));	\
  res; })

/* outportb() =  Write a byte to an I/O port */
#define outportb(Port,Data) {			\
  if (__builtin_constant_p(Port)) 		\
    asm volatile(" outb %%al, %1\n" 		\
		 : : "a" (Data),  "i" (Port));	\
  else						\
    asm volatile(" outb %%al,%%dx\n"		\
		 : : "a" (Data), "d" (Port)); }


/* inportw() = Get a word from an I/O port */
inline static unsigned short
inportw(int port) {
  register unsigned short res;
  asm volatile("inw %%dx,%%ax\n\t"
	       : "=a" (res) : "d" (port));
  return(res); }

/* outportw() = Write a word to an I/O port */
inline static void
outportw(int port, unsigned short data) {
  asm volatile("outw %%ax,%%dx\n\t"
	       : /* No output */  : "a" (data), "d" (port)); }


inline static void
repinsw(int port, void* va, int count) {
  asm volatile("rep\n\t"
	       "insw" 
	       :
	       : "d" (port), "D" (va), "c" (count)
	       : "di"); }  

inline static void
repoutsw(int port, void* va, int count) {
  asm volatile("rep\n\t"
	       "outsw" 
	       :
	       : "d" (port), "D" (va), "c" (count)
	       : "di"); }

inline static void
repinsl(int port, void* va, int count) {
  asm volatile("rep\n\t"
	       "insl" 
	       :
	       : "d" (port), "D" (va), "c" (count)
	       : "di"); }  

inline static void
repoutsl(int port, void* va, int count) {
  asm volatile("rep\n\t"
	       "outsl" 
	       :
	       : "d" (port), "D" (va), "c" (count)
	       : "di"); }


extern inline char * strcpy(char * dest,const char *src)
{
  register char *tmp= (char *)dest;
  register char dummy;
  asm volatile("\n1:\t"
	       "movb (%0),%b2\n\t"
	       "incl %0\n\t"
	       "movb %b2,(%1)\n\t"
	       "incl %1\n\t"
	       "testb %b2,%b2\n\t"
	       "jne 1b"
	       :"=r" (src), "=r" (tmp), "=q" (dummy)
	       :"0" (src), "1" (tmp)
	       :"memory");
  return dest;
}


void memset(void*, char, unsigned int);

extern inline void memset(void *s, char c,unsigned int count)
{
  asm volatile(
	       "cld\n\t"
	       "rep\n\t"
	       "stosb"
	       : /* no output */
	       :"a" (c),"D" (s),"c" (count)
	       :"cx","di","memory");
}

#define strlen __builtin_strlen
#define memcpy __builtin_memcpy
#define memcmp __builtin_memcmp
#define sti() asm volatile ("sti": : :"memory")
#define cli() asm volatile ("cli": : :"memory")
/* define somehow reentrant atomic blocks with 
   {i=int_saved(); atomic_do(); int_restore(i); } */

inline static int int_saved(void) {
  register int ret;
  asm volatile("pushfl\n\t"
	       "popl %0\n\t"
	       "cli"
	       : "=r" (ret) :	);
  return (ret & 0x200);
}

inline static void int_restore(int l) {
  if(l) sti();
}

#define UGLY_BUSY_DELAY(N) {int i; \
  for (i=(N)<<20; i>0; i--) asm volatile "nop;nop;nop") }

#define iret() asm volatile ("iret": : :"memory")

/* enable/disable IRQ line */
void enable_irq(int irqnb);
void disable_irq(int irqnb);

/* some low level (descriptors, interrupts) taken verbatim from Bart
   Sekura's TINOS */
#define BIT(n) (1<<(n))

typedef unsigned short ushort;
typedef unsigned char uchar;
typedef unsigned long ulong;

struct gate_desc {
	ushort   offset_0;
   	ushort   selector;
  	ushort   type;
  	ushort   offset_16;
};

struct sys_desc {
   	ushort   limit_0_15;
   	ushort   base_0_15;
   	uchar    base_16_23;
   	uchar    access;
   	uchar    gd_limit_16_19;
   	uchar    base_24_31;
};

/* these are bit settings for field access in sys_desc */
#define A_PRESENT	BIT(7)
/* descriptor priviledge level - bits 6-5 */
#define A_DPL_0		(0)
#define A_DPL_1		BIT(5)	
#define A_DPL_2		BIT(6)
#define A_DPL_3		BIT(6) | BIT(5)
/* descriptor type and access - bit 4 and bits 3-1 */
#define A_DATA		BIT(4)	
#define A_CODE		BIT(4) | BIT(3)
#define A_DATA_WRITE	BIT(4) | BIT(1)
#define A_CODE_READ	BIT(4) | BIT(3) | BIT(1)

#define A_ACCESSED	BIT(0)

/* these bits are for field gd_limit_16_19 */
#define SEG_PAGED	BIT(7)
#define SEG_32		BIT(6)



/* gdt selectors */
#define GDTSEL_KCODE	(0x08)
#define GDTSEL_KDATA	(0x10)
#define GDTSEL_UCODE	(0x18)
#define GDTSEL_UDATA	(0x20)
#define GDTSEL_KTSS	(0x28)

#define USER_CSEG	(GDTSEL_UCODE | 0x03)
#define	USER_DSEG	(GDTSEL_UDATA | 0x03)

#define	KERN_BASE	(0x00000000)
#define USER_BASE	(0x40000000)

/* pseudo descriptor is used to load descriptor registers
 * in the following inline functions */
struct pseudo_desc {
  ushort	pad;
  ushort	len;
  ulong	base;
};

/* standard frame (taken from Tinos) used in low level interrupt routines */

struct frame_st {
  ulong esds;	/* these we put lastly */
  ulong edi;	/* these are our pushl's */
  ulong esi; 
  ulong ebp; 
  ulong ebx;
  ulong edx; 
  ulong ecx; 
  ulong eax;  
  ulong err;	/* this is either zero or some error code */
  ulong eip;	/* this is extra bonus from Intel */
  ulong ecs;
  ulong eflags;
  ulong esp;
  ulong ess;
};


/* the screen low level routines */
#define SCR_ATTR_NORM 0x07

/* clear the screen */
void scr_clear(void); 

/* show a string on a screen - no characters interpretation; don't
   wrap at end of line! */
void scr_raw_show(char *s, int len);

/* show a C-null-terminated-string on a screen - no characters
   interpretation; don't wrap at end of line! */
void scr_c_show(char *s);

/* update the hardware cursor */
void scr_update_cursor(void);

/* new line on screen, with scrolling if needed */
void scr_nl(void);

/* newline if not in first column */
void scr_maybenl(void);

/* show a single character */
void scr_char_show(char c);


/* show a signed integer on the screen in decimal */
void scr_show_decint(int i);

/* show an unsigned word on the screen in hexadecimal */
void scr_show_hexword(unsigned w);

#define scrs_sn(Msg) {scr_c_show(Msg); scr_nl();}
#define scrs_s(Msg) {scr_c_show(Msg);}

#define scrs_ssn(Msg1,Msg2)  {scr_c_show(Msg1); scr_c_show(Msg2); scr_nl();}
#define scrs_sin(Msg,Nb) {scr_c_show(Msg); scr_show_decint(Nb); scr_nl();}
#define scrs_shn(Msg,Nb) {scr_c_show(Msg); scr_show_hexword(Nb); scr_nl();}
#define scrs_ss(Msg1,Msg2)  {scr_c_show(Msg1); scr_c_show(Msg2);}
#define scrs_si(Msg,Nb) {scr_c_show(Msg); scr_show_decint(Nb);}
#define scrs_sh(Msg,Nb) {scr_c_show(Msg); scr_show_hexword(Nb);}

#define scrs_ssin(Msg1,Msg2)  {scr_c_show(Msg1); scr_c_show(Msg2); scr_nl();}
#define scrs_siin(Msg,Nb) {scr_c_show(Msg); scr_show_decint(Nb); scr_nl();}
#define scrs_shin(Msg,Nb) {scr_c_show(Msg); scr_show_hexword(Nb); scr_nl();}
#define scrs_ssi(Msg1,Msg2)  {scr_c_show(Msg1); scr_c_show(Msg2);}
#define scrs_sii(Msg,Nb) {scr_c_show(Msg); scr_show_decint(Nb);}

void panic_at(char* fil, int lin, char *msg);
void panic_at_int(char* fil, int lin, char *msg, int nb);
void panic_at_word(char* fil, int lin, char *msg, unsigned nb);

#define panic(Msg) panic_at(__FILE__,__LINE__,(Msg))
#define panic_int(Msg,Nb) panic_at_int(__FILE__,__LINE__,(Msg),(Nb))
#define panic_word(Msg,Nb)  panic_at_word(__FILE__,__LINE__,(Msg),(Nb))

/* set cursor position & attribute */
void scr_setpos(int x, int y);
void scr_setattr(int mod);

/* get cursor position */
int scr_posx(void);
int scr_posy(void);


/**************** requests ****************/
typedef int reqid_t;		/* should be >0 (<=0 indicate error) */

/* create a request from a facility - return <=0 on error */
reqid_t request_create(unsigned char typ, void*data, unsigned priv);

/* called with disabled interrupts from facility or interrupt handler
   - return 0 iff ok */
int request_end(reqid_t rqid, unsigned res);


/* wait a given request - return 0 iff ok and set *pres to result */
int request_wait(reqid_t rqid, unsigned *pres);

/* maximum length of request arrays */
#define REQARRAY_MAXLEN           64

/* wait for one of a given array of requests */
reqid_t request_wait_some(reqid_t rqidtable[], int len, unsigned *pres);

/* wait for all of a given array of requests - return 0 iff ok */
int request_wait_all(reqid_t rqidtable[], unsigned restable[], int len);

/* wait for any arbitrary pending request */
reqid_t request_wait_any(unsigned *pres);

/* test if request is completed without waiting */
int request_completed(reqid_t rqid, unsigned *pres);

/* ask for a request to be signaled asynchronously */
int request_signal(reqid_t rqid, void*data, void (*handler)(reqid_t,void*));

/* request types (>0) */
#define REQTYPE_KEYBOARD       1
#define REQTYPE_TIMER          2

/* make a keyboard request - its result is the keysym [mostly unicode]*/
reqid_t keyboard_request(void);


void kut_main(void);
void init_memory(void);
void init_interrupts(void);
void init_keyboard(void);

/* called from kutinter.c but coded in kutkeyb.c */
void handle_irq1keyboard(unsigned stack);
#endif /*  __KUT_INCLUDED__ */

/**** rcslog ****
 * $Log: kut.h,v $
 * Revision 0.8  1998/08/30 09:19:02  basile
 * added struct frame_st (from Tinos)
 *
 * Revision 0.7  1998/08/24 20:13:48  basile
 * added an inlined memset
 *
 * Revision 0.6  1998/08/23 11:04:31  basile
 * optimize inportb & outportb for compiletime constant ports
 *
 * Revision 0.5  1998/08/22 18:49:42  basile
 * added request stuff
 *
 * Revision 0.4  1998/07/29 06:18:13  basile
 * added keybord interrupt
 *
 * Revision 0.3  1998/07/18 22:05:02  basile
 * added screen stuff
 * added inlined I/O port routines
 *
 * Revision 0.2  1998/07/16 17:57:55  basile
 * ok
 *
 * Revision 0.1  1998/07/16 17:54:17  basile
 **** endlog *****/

/***eof $Header: /home/basile/KUT/RCS/kut.h,v 0.8 1998/08/30 09:19:02 basile Exp $****/
