/* file KUT/kutinter.c */
/* Time-stamp: <1998 Augu 30 Sun 19:26:45 CEST Basile.Starynkevitch@wanadoo.fr> */
/* rcsid: $Id: kutinter.c,v 0.9 1998/09/02 20:28:44 basile Exp $ */

#include "kut.h"

const char kutinter_ident[] = 
 "@(#)$Id: kutinter.c,v 0.9 1998/09/02 20:28:44 basile Exp $ "
 "<" __DATE__ "@" __TIME__ ">";

inline static void
load_idt(struct pseudo_desc* pdesc) {
  asm volatile ("lidt %0" : : "m" (pdesc->len));
}


/* trap types */
#define     TRAP_GATE      (0x8F00)
#define     IRQ_GATE       (0x8E00)
#define     USER_TRAP_GATE (0xEF00)


#define IDT_SIZE 256
struct gate_desc idt[IDT_SIZE] = {
#define IDT_ENTRY(Handler,Type) {((Handler)&0xffff), GDTSEL_KCODE, (Type), ((Handler)>>16)}
  /* unfortunately we can't use IDT_ENTRY because the linker don't
     know how to compute complex expressions at link time */
  IDT_ENTRY(0,0)		/* fake entry to avoid .bss section! */
};

/* interrupts and traps */

 
inline static void setup_gate(int i, ulong p_hand, ushort type) {
  idt[i].offset_0 = p_hand;
  idt[i].selector = GDTSEL_KCODE;
  idt[i].type = type;
  idt[i].offset_16 = (p_hand >> 16);
}

/* declare the interrupt handlers - declare handle_* C and int_*
   assembly routines */

#define INTHANDLER(Nam,Rank)                        \
 void handle_##Nam(unsigned);  void int_##Nam(void);

#define INTERRHANDLER(Nam,Rank)                     \
 void handle_##Nam(unsigned);  void int_##Nam(void);

#define IRQAHANDLER(Nam,Rank)                     \
 void handle_##Nam(unsigned); void int_##Nam(void);

#define IRQBHANDLER(Nam,Rank)                     \
 void handle_##Nam(unsigned);  void int_##Nam(void);

#include "interr.tab"

#undef INTHANDLER
#undef INTERRHANDLER
#undef IRQAHANDLER
#undef IRQBHANDLER

static struct pseudo_desc idt_desc = {
  /* pad= */ 0, /* len= */sizeof(idt) - 1, /* base= */(ulong)&idt };


static void timer_init(void);

void handle_divide(unsigned stack) {
  struct frame_st *f = (struct frame_st*)&stack;
  panic_word("unexpected zerodivide from", f->eip);
}

void handle_debug(unsigned stack) {
  struct frame_st *f = (struct frame_st*)&stack;
  panic_word("unexpected debug from", f->eip);
}

void handle_nmi(unsigned stack) {
  struct frame_st *f = (struct frame_st*)&stack;
  panic_word("unexpected non maskable interrupt from", f->eip);
}

void handle_breakpoint(unsigned stack) {
  struct frame_st *f = (struct frame_st*)&stack;
  scrs_sh("breakpoint from ", f->eip);
  scrs_sh("(hex) err=", f->err);
  scrs_sn("(hex).");
  /* panic_word("unexpected breakpoint from", f->eip); */
}

void handle_overflow(unsigned stack) {
  struct frame_st *f = (struct frame_st*)&stack;
  panic_word("unexpected overflow from", f->eip);
}

void handle_bound(unsigned stack) {
  struct frame_st *f = (struct frame_st*)&stack;
  panic_word("unexpected bound error from", f->eip);
}

void handle_badopcode(unsigned stack) {
  struct frame_st *f = (struct frame_st*)&stack;
  panic_word("unexpected bad opcode from", f->eip);
}

void handle_floatunavailable(unsigned stack) {
  struct frame_st *f = (struct frame_st*)&stack;
  panic_word("unexpected float unavailable from", f->eip);
}

void handle_doublefault(unsigned stack) {
  struct frame_st *f = (struct frame_st*)&stack;
  panic_word("double fault from", f->eip);
}

void handle_coprocessorsegment(unsigned stack) {
  struct frame_st *f = (struct frame_st*)&stack;
  panic_word("coprocessor segment from", f->eip);
}

void handle_invalidtss(unsigned stack) {
  struct frame_st *f = (struct frame_st*)&stack;
  panic_word("invalid TSS from", f->eip);
}

void handle_segmentnotpresent(unsigned stack) {
  struct frame_st *f = (struct frame_st*)&stack;
  panic_word("segment not present from", f->eip);
}

void handle_stackfault(unsigned stack) {
  struct frame_st *f = (struct frame_st*)&stack;
  panic_word("stack fault from", f->eip);
}

void handle_generalprotection(unsigned stack) {
  struct frame_st *f = (struct frame_st*)&stack;
  panic_word("general protection fault from", f->eip);
}

void handle_pagefault(unsigned stack) {
  struct frame_st *f = (struct frame_st*)&stack;
  panic_word("page fault at", f->eip);
}

void handle_intelresv15(unsigned stack) {
  struct frame_st *f = (struct frame_st*)&stack;
  panic_word("intel reserved 15 fault at", f->eip);
}

void handle_floaterror(unsigned stack) {
  struct frame_st *f = (struct frame_st*)&stack;
  /* the faulting instruction is in FPU, it is not F->EIP! */
  panic_word("float error detected at", f->eip);
}

void handle_alignment(unsigned stack) {
  struct frame_st *f = (struct frame_st*)&stack;
  /* only happens in user mode! */
  panic_word("alignment user fault at", f->eip);
}

void handle_machinecheck(unsigned stack) {
  struct frame_st *f = (struct frame_st*)&stack;
  panic_word("machine check fault at", f->eip);
}

unsigned long timer_ticks;
unsigned long hightimer_count;





void handle_irq0timer(unsigned stack) {
  struct frame_st *f = (struct frame_st*)&stack;
  int y;
  timer_ticks++;
  if ((timer_ticks & 0xffff) == 0) 
    hightimer_count++;
#ifdef KUT_DEBUG
  y = scr_posy();
  scr_setpos(0,2);
  scrs_si("irq0 timer !#", timer_ticks);
  scrs_sh(" from IP=0x", f->eip);
  scr_setpos(0,y);
#endif /*KUT_DEBUG*/
}


void handle_irq2cascadenever(unsigned stack) {
  struct frame_st *f = (struct frame_st*)&stack;
  panic_word("unexpected irq2 from ", f->eip);
}

void handle_irq3unused(unsigned stack) {
  struct frame_st *f = (struct frame_st*)&stack;
  panic_word("unexpected irq3unused from ", f->eip);
}

void handle_irq4serial(unsigned stack) {
  struct frame_st *f = (struct frame_st*)&stack;
  panic_word("unexpected irq4serial from ", f->eip);
}

void handle_irq14ide0(unsigned stack) {
  struct frame_st *f = (struct frame_st*)&stack;
  panic_word("unexpected irq14 IDE0 from ", f->eip);
}

void handle_irq15ide1(unsigned stack) {
  struct frame_st *f = (struct frame_st*)&stack;
  panic_word("unexpected irq15 IDE1 from ", f->eip);
}
/******************** 8259 interrupt control units ********************/
/* most code taken from TINOS */

/* 8259 ICU */
/* Base addresses of the two 8259 units */
#define ICU0 (0x20)
#define ICU1 (0xA0)

/* End Of Interrupt flag */
#define EOI_FLAG (0x20)

/* number of interrupts used by CPU ie first rank of IRQ interrupt vector */
#define CPUIDT 32

/* programmable interrupt controler ports */

#define MASTER_PIC_CMD		(0x20)
#define MASTER_PIC_INTMASK	(0x21)

#define SECOND_PIC_CMD		(0xA0)
#define SECOND_PIC_INTMASK	(0xA1)

/* initialization command words */

#define ICW1_NEED_ICW4		BIT(0)
#define ICW1_NO_ICW4		(0)
#define ICW1_SINGLE		BIT(1)
#define ICW1_CASCADING		(0)
#define ICW1_4_BYTE_INTVEC	BIT(2)
#define ICW1_8_BYTE_INTVEC	(0)
#define ICW1_LEVEL_TRIG		BIT(3)
#define ICW1_EDGE_TRIG		(0)
#define ICW1_BASE		BIT(4)

#define ICW2_INTVEC_MASK	(BIT(7)|BIT(6)|BIT(5)|BIT(4)|BIT(3))

#define ICW3_INT_0_SLAVE	BIT(0)
#define ICW3_INT_1_SLAVE	BIT(1)
#define ICW3_INT_2_SLAVE	BIT(2)
#define ICW3_INT_3_SLAVE	BIT(3)
#define ICW3_INT_4_SLAVE	BIT(4)
#define ICW3_INT_5_SLAVE	BIT(5)
#define ICW3_INT_6_SLAVE	BIT(6)
#define ICW3_INT_7_SLAVE	BIT(7)

#define ICW3_SLAVE_ON_1		(0x01)
#define ICW3_SLAVE_ON_2		(0x02)

#define ICW4_80x86_MODE		BIT(0)
#define ICW4_8085_MODE		(0)
#define ICW4_AUTO_EOI		BIT(1)
#define ICW4_NORMAL_EOI		(0)
#define ICW4_SFN_MODE		BIT(4)
#define ICW4_SEQUENTIAL		(0)
#define ICW4_BUF_SLAVE		BIT(3)
#define ICW4_BUF_MASTER		BIT(2) | BIT(3)

#define OCW1_MASK_OFF_ALL	(0xFF)

#define OCW2_NON_SPEC_EOI	BIT(5)

 /*  8259 interrupt control units initialization  */
static void icu_init(void) {
  /* initialize 8259's */
#define MASTER_INT_OFFSET      0x20
#define SLAVE_INT_OFFSET       0x28

  outportb(MASTER_PIC_CMD, ICW1_BASE | ICW1_NEED_ICW4);
  outportb(MASTER_PIC_INTMASK, ICW2_INTVEC_MASK & MASTER_INT_OFFSET);
  outportb(MASTER_PIC_INTMASK, ICW3_INT_2_SLAVE);
  outportb(MASTER_PIC_INTMASK, ICW4_80x86_MODE);
  outportb(MASTER_PIC_INTMASK, OCW1_MASK_OFF_ALL);
  outportb(MASTER_PIC_CMD, OCW2_NON_SPEC_EOI);
	
  outportb(SECOND_PIC_CMD, ICW1_BASE | ICW1_NEED_ICW4);
  outportb(SECOND_PIC_INTMASK, ICW2_INTVEC_MASK & SLAVE_INT_OFFSET);
  outportb(SECOND_PIC_INTMASK, ICW3_SLAVE_ON_2);
  outportb(SECOND_PIC_INTMASK, ICW4_80x86_MODE);
  outportb(SECOND_PIC_INTMASK, OCW1_MASK_OFF_ALL);
  outportb(SECOND_PIC_CMD, OCW2_NON_SPEC_EOI);

}

/* set_irq_mask() - set the interrupt mask */
inline static void set_irq_mask(unsigned short mask) {
  outportb(MASTER_PIC_INTMASK, mask & 0xff);
  outportb(SECOND_PIC_INTMASK, (mask >> 8) & 0xff);
}



#define SLAVE_IRQ	8
#define MASTER_SLAVE	2

static unsigned short irq_mask= 0xffff;

void enable_irq(int irqnb) {
#ifdef KUT_DEBUG
  scrs_sin("enable_irq #", irqnb);
#endif /*KUT_DEBUG*/
  irq_mask &= ~(1<<irqnb);
  if(irqnb >= SLAVE_IRQ) 
    irq_mask &= ~(1 << MASTER_SLAVE);
  set_irq_mask(irq_mask);
}

void disable_irq(int irqnb) {
#ifdef KUT_DEBUG
  scrs_sin("disable_irq #", irqnb);
#endif /*KUT_DEBUG*/
  irq_mask |= (1<<irqnb);
  if(irqnb >= SLAVE_IRQ) 
    irq_mask |= (1 << MASTER_SLAVE);
  set_irq_mask(irq_mask);
}

/*****************  8254 programmable interrupt timer ****************/

/* 8254 programmable interrupt timer */
#define PITCTR0_PORT	(0x40)		/* counter 0 */
#define PITCTR1_PORT	(0x41)		/* counter 1 */
#define PITCTR2_PORT	(0x42)		/* counter 2 */

#define PITCTL_PORT	(0x43)		/* control port */
#define PITAUX_PORT	(0x61)		/* auxiliary port */

/* 8254 commands */

/* timer 0 */
#define PIT_C0		(0x00)		/* select counter 0 */
#define PIT_LOADMODE	(0x30)		/* load LSB followed by MSB */
#define PIT_NDIVMODE	(0x04)		/* divide by N counter */
#define PIT_SQUAREMODE	(0x06)		/* square-wave mode */

/* timer 1 */
#define PIT_C1		(0x40)		/* select counter 1 */
#define PIT_READMODE	(0x30)		/* read or load LSB, MSB */
#define PIT_RATEMODE	(0x06)		/* sqaure-wave mode for USART */

/* auxiliary control port for timer 2 */
#define PITAUX_GATE2	(0x01)		/* aux port, PIT gate 2 input */
#define PITAUX_OUT2	(0x02)		/* aux port, PIT clock out 2 enable */

#define CLKNUM	(1193167)

#define HZ	(2)

/*** FIXME! something don't work as expected. It seems that the timer
   irq handler is called all the time (every 10ms with a big busy
   delay in the timer handler!)*/
static void
timer_init(void) {
#define clock (CLKNUM/HZ)
  outportb(PITCTL_PORT, PIT_C0 | PIT_SQUAREMODE | PIT_LOADMODE);
  outportb(PITCTR0_PORT, clock & 0x00FF);
  outportb(PITCTR0_PORT, (clock & 0xFF00) >> 8);
#ifdef KUT_DEBUG
  scrs_sin("timer_init DO enable irq clock=", clock);
#endif 
  enable_irq(0);  
}


/***************** initializer called from main *****************/
void init_interrupts(void)  {
  /* install the interrupt handlers */
#define INTHANDLER(Nam,Rank) setup_gate(Rank,(unsigned)&int_##Nam,IRQ_GATE);
#define INTERRHANDLER(Nam,Rank) setup_gate(Rank,(unsigned)&int_##Nam,IRQ_GATE);
#define IRQAHANDLER(Nam,Rank) setup_gate(Rank,(unsigned)&int_##Nam,IRQ_GATE);
#define IRQBHANDLER(Nam,Rank) setup_gate(Rank,(unsigned)&int_##Nam,IRQ_GATE);
#include "interr.tab"
#undef INTHANDLER
#undef INTERRHANDLER
#undef IRQAHANDLER
#undef IRQBHANDLER
  /* load the interrupt descriptor table */
  load_idt(&idt_desc);
  /* initialize the timer */
  if (0) timer_init();
  scrs_sn("no timer");
  /* initialize the ICU - but IRQ are still all disabled */
  icu_init();
  /* enabling 1 2 3 */
  scrs_sn("enabling IRQ 1 2 3 4");
  enable_irq(1);  enable_irq(2);  enable_irq(3); enable_irq(4);
  scrs_shn("irq_mask=", irq_mask);
  set_irq_mask(irq_mask);
#ifdef KUT_DEBUG
  scrs_sn("[kutinter.c $Revision: 0.9 $] interrupts initialized!");
#endif /*KUT_DEBUG*/
}

/*** rcslog
 ** $Log: kutinter.c,v $
 ** Revision 0.9  1998/09/02 20:28:44  basile
 ** still don't work
 **
 ** Revision 0.8  1998/08/30 10:24:05  basile
 ** VSTa interrupt code doesn't work well - trying Tinos...
 **
 ** Revision 0.7  1998/08/30 09:19:37  basile
 ** use Tinos struct frame_st
 **
 ** Revision 0.6  1998/08/24 20:15:06  basile
 ** dont work really well. The timer IRQ0 handler is called all the time!
 **
 ** Revision 0.5  1998/08/23 11:05:25  basile
 ** added ICU but don't get any keyboard interrupts
 **
 ** Revision 0.4  1998/08/22 21:20:20  basile
 ** should initialize ICU
 **
 ** Revision 0.3  1998/07/19 20:27:45  basile
 ** added pseudo timer routines
 **
 ** Revision 0.2  1998/07/18 22:05:46  basile
 ** added some interrupt handlers
 **
 ** Revision 0.1  1998/07/16 17:55:16  basile
 **
 *** endlog ****/

/* eof $Header: /home/basile/KUT/RCS/kutinter.c,v 0.9 1998/09/02 20:28:44 basile Exp $ */



