/* file KUT/kutkeyb.c - keyboard driver */
/* Time-stamp: <1998 Augu 30 Sun 18:30:05 CEST Basile.Starynkevitch@wanadoo.fr> */
/* rcsid: $Id: kutkeyb.c,v 0.8 1998/09/02 20:29:08 basile Exp $ */

#include "kut.h"

const char kutkeyb_ident[] = 
 "@(#)$Id: kutkeyb.c,v 0.8 1998/09/02 20:29:08 basile Exp $ "
 "<" __DATE__ "@" __TIME__ ">";

/* some stuff taken from Tinos by Bart Sekura who borrowed from others
   (Vsta, Linux); some stuff taken from Linux-2.1 */

/* keyboard use IRQ1 */
#define KB_IRQ                  1

#define KB_STATUS_REG		(0x64)

#define KBS_OUT_DATA		(0x01)
#define KBS_IN_DATA		(0x02)
#define KBS_TX_TIMEOUT		(0x20)
#define KBS_RECV_TIMEOUT	(0x40)
#define KBS_PAR_EVEN		(0x80)

#define KB_CMD_REG		(0x64)

#define KBC_SELF_TEST		(0xAA)
#define KBC_IF_TEST		(0xAB)
#define KBC_IF_DISABLE		(0xAD)
#define KBC_IF_ENABLE		(0xAE)

#define KB_DATA_REG		(0x60)

#define KB_ENABLE		(0xF4)
#define KB_ACK			(0xFA)
#define KB_RESEND		(0xFE)
#define KB_LEDS                 (0xED)

#define	KB_CTL_REG		(0x61)

#define KBLED_CAPS           04
#define KBLED_NUM            02
#define KBLED_SCROLL         01

static void create_keybuf(void);

/* read keyboard and always check status if there's some stuff for us */
static inline int kb_read(void) {
  int c = 100000;
  while(c--) {
    int status = inportb(KB_STATUS_REG);
    if(!(status & KBS_OUT_DATA))
      continue;  
    return (inportb(KB_DATA_REG) & 0xFF);
  }
  return (-1);
}

/* write some stuff to keyboard busy-wait spinning while it's not
 * ready */
static inline void 
kb_write(int port, int data) {
  int status;
  do {
    status = inportb(KB_STATUS_REG);
    asm volatile ("nop");
  } while(status & KBS_IN_DATA);
  outportb(port, data);
}

static unsigned kbd_leds;

static void update_kbd_leds(void) {
  kb_write(KB_DATA_REG, KB_LEDS);
  kb_write(KB_DATA_REG, kbd_leds & 07);
}

static inline void show_caps_led(int f) {
  if (f) kbd_leds |= KBLED_CAPS; else kbd_leds &= (~KBLED_CAPS);
  update_kbd_leds();
}

static inline void show_nums_led(int f) {
  if (f) kbd_leds |= KBLED_NUM; else kbd_leds &= (~KBLED_NUM);
  update_kbd_leds();
}
static inline void show_scroll_led(int f) {
  if (f) kbd_leds |= KBLED_SCROLL; else kbd_leds &= (~KBLED_SCROLL);
  update_kbd_leds();
}

/* following stuff taken from linux-2.1 drivers/char/pc_keyb.c */

/*
 * Translation of escaped scancodes to keycodes.
 * This is now user-settable.
 * The keycodes 1-88,96-111,119 are fairly standard, and
 * should probably not be changed - changing might confuse X.
 *
 * For 1-88 keycode equals scancode.
 */
#define SC_LIM 89

#define E0_KPENTER 96
#define E0_RCTRL   97
#define E0_KPSLASH 98
#define E0_PRSCR   99
#define E0_RALT    100
#define E0_BREAK   101  /* (control-pause) */
#define E0_HOME    102
#define E0_UP      103
#define E0_PGUP    104
#define E0_LEFT    105
#define E0_RIGHT   106
#define E0_END     107
#define E0_DOWN    108
#define E0_PGDN    109
#define E0_INS     110
#define E0_DEL     111

#define E1_PAUSE   119



/* BTC */
#define E0_MACRO   112
/* LK450 */
#define E0_F13     113
#define E0_F14     114
#define E0_HELP    115
#define E0_DO      116
#define E0_F17     117
#define E0_KPMINPLUS 118
/*
 * My OmniKey generates e0 4c for  the "OMNI" key and the
 * right alt key does nada. [kkoller@nyx10.cs.du.edu]
 */
#define E0_OK	124
/*
 * New microsoft keyboard is rumoured to have
 * e0 5b (left window button), e0 5c (right window button),
 * e0 5d (menu button). [or: LBANNER, RBANNER, RMENU]
 * [or: Windows_L, Windows_R, TaskMan]
 */
#define E0_MSLW	125
#define E0_MSRW	126
#define E0_MSTM	127

static unsigned char e0_keys[128] = {
  0, 0, 0, 0, 0, 0, 0, 0,			      /* 0x00-0x07 */
  0, 0, 0, 0, 0, 0, 0, 0,			      /* 0x08-0x0f */
  0, 0, 0, 0, 0, 0, 0, 0,			      /* 0x10-0x17 */
  0, 0, 0, 0, E0_KPENTER, E0_RCTRL, 0, 0,	      /* 0x18-0x1f */
  0, 0, 0, 0, 0, 0, 0, 0,			      /* 0x20-0x27 */
  0, 0, 0, 0, 0, 0, 0, 0,			      /* 0x28-0x2f */
  0, 0, 0, 0, 0, E0_KPSLASH, 0, E0_PRSCR,	      /* 0x30-0x37 */
  E0_RALT, 0, 0, 0, 0, E0_F13, E0_F14, E0_HELP,	      /* 0x38-0x3f */
  E0_DO, E0_F17, 0, 0, 0, 0, E0_BREAK, E0_HOME,	      /* 0x40-0x47 */
  E0_UP, E0_PGUP, 0, E0_LEFT, E0_OK, E0_RIGHT, E0_KPMINPLUS, E0_END,/* 0x48-0x4f */
  E0_DOWN, E0_PGDN, E0_INS, E0_DEL, 0, 0, 0, 0,	      /* 0x50-0x57 */
  0, 0, 0, E0_MSLW, E0_MSRW, E0_MSTM, 0, 0,	      /* 0x58-0x5f */
  0, 0, 0, 0, 0, 0, 0, 0,			      /* 0x60-0x67 */
  0, 0, 0, 0, 0, 0, 0, E0_MACRO,		      /* 0x68-0x6f */
  0, 0, 0, 0, 0, 0, 0, 0,			      /* 0x70-0x77 */
  0, 0, 0, 0, 0, 0, 0, 0			      /* 0x78-0x7f */
};


static unsigned int prev_scancode = 0;   /* remember E0, E1 */

static inline int pckbd_pretranslate(unsigned char scancode)
{
  if (scancode == 0xff) {
    /* in scancode mode 1, my ESC key generates 0xff */
    prev_scancode = 0;
    return 0;
  }

  if (scancode == 0xe0 || scancode == 0xe1) {
    prev_scancode = scancode;
    return 0;
  }
  return 1;
}

static inline int pckbd_translate(unsigned char scancode, unsigned char *keycode)
{
  if (prev_scancode) {
    /*
     * usually it will be 0xe0, but a Pause key generates
     * e1 1d 45 e1 9d c5 when pressed, and nothing when released
     */
    if (prev_scancode != 0xe0) {
      if (prev_scancode == 0xe1 && scancode == 0x1d) {
	prev_scancode = 0x100;
	return 0;
      } else if (prev_scancode == 0x100 && scancode == 0x45) {
	*keycode = E1_PAUSE;
	prev_scancode = 0;
      } else {
	prev_scancode = 0;
	return 0;
      }
    } else {
      prev_scancode = 0;
      /*
       *  The keyboard maintains its own internal caps lock and
       *  num lock statuses. In caps lock mode E0 AA precedes make
       *  code and E0 2A follows break code. In num lock mode,
       *  E0 2A precedes make code and E0 AA follows break code.
       *  We do our own book-keeping, so we will just ignore these.
       */
      /*
       *  For my keyboard there is no caps lock mode, but there are
       *  both Shift-L and Shift-R modes. The former mode generates
       *  E0 2A / E0 AA pairs, the latter E0 B6 / E0 36 pairs.
       *  So, we should also ignore the latter. - aeb@cwi.nl
       */
      if (scancode == 0x2a || scancode == 0x36)
	return 0;

      if (e0_keys[scancode])
	*keycode = e0_keys[scancode];
      else 
	return 0;
    }
  } else if (scancode >= SC_LIM) 
    return 0;
  else
    *keycode = scancode;
  return 1;
}

static void emit_keysym(int symnum);

/** this included file has the 
     dispatch_keycode(unsigned code,unsigned pressed) routine, 
   and defines many keysyms; this included file is generated from
   a keyboard map by ut_kbdmap.ml utility [coded in Caml] */

#include "_kbddisp.hc"

void handle_irq1keyboard(unsigned stack) {
  struct frame_st *f = (struct frame_st*)&stack;
  unsigned char scancode, keycode, up_flag;
  int status;
#ifdef KUT_DEBUG
  int x,y;
  x = scr_posx();
  y = scr_posy();
  scr_setpos(5,10);
  scrs_sn("kbd irq ");
  scr_setpos(x,y);
#endif /*KUT_DEBUG*/
  for (;;) {
    scancode = kb_read();
    asm volatile ("nop;nop;nop;nop"); /* microdelay */
#ifdef KUT_DEBUG
    scrs_sin("kbd got raw scancode", scancode);
#endif /*KUT_DEBUG*/
    if (pckbd_pretranslate(scancode)) {
      up_flag = (scancode & 0200);
      scancode &= 0x7f;
      keycode = 0xff;
      if (pckbd_translate(scancode, &keycode)) {
	dispatch_keycode(keycode,up_flag);
      };
    }
    status = inportb(KB_STATUS_REG);
    if(!(status & KBS_OUT_DATA))
      return; 
  };
}

#ifndef KEYSYM_BUFLEN
#define KEYSYM_BUFLEN 64
#endif /*KEYSYM_BUFLEN*/

/* all writes into the Keybuf should be with interrupt disabled! */ 
static 
struct keybuf_st {
  unsigned short kb_buf[KEYSYM_BUFLEN];
  short kb_count;			/* number of keysyms */
  short kb_rkin;		/* rank of emitted keysym */
  short kb_rkout;		/* rank of read keysym */
} Keybuf;


static void create_keybuf(void) {
  int i;
  for (i=0; i<KEYSYM_BUFLEN; i++) Keybuf.kb_buf[i] = 0;
  Keybuf.kb_count = 0; 
  Keybuf.kb_rkin = Keybuf.kb_rkout = 0;
}

/* we have at most one pending keyboard request */
static  reqid_t kbd_rqid;

/* called from dispatch_keycode (which is called from interrupt
   handler, with disabled interrupts) */
static void emit_keysym(int symnum) {
  /* ignore keysym if keybuffer full */
  if (Keybuf.kb_count >= KEYSYM_BUFLEN-1) 
    return;
  Keybuf.kb_count++;
  Keybuf.kb_buf[Keybuf.kb_rkin] = symnum;
  Keybuf.kb_rkin = (Keybuf.kb_rkin+1) % KEYSYM_BUFLEN;
  if (kbd_rqid) {
    unsigned keysym = Keybuf.kb_buf[Keybuf.kb_rkout];
    reqid_t rqid = kbd_rqid;
    kbd_rqid = 0;
    Keybuf.kb_rkout = (Keybuf.kb_rkout+1) % KEYSYM_BUFLEN;
    Keybuf.kb_count--;
    (void) request_end(rqid, keysym);
  };
}

/* keyboard facility */

reqid_t keyboard_request(void) {
  int intlev;
  reqid_t rqid=0;
  intlev = int_saved();
  /* at most one pending keyboard request! */
  if (kbd_rqid!=0) goto end;
  kbd_rqid = rqid = request_create(REQTYPE_KEYBOARD, (void*)0, 0);
  /* if already recieved keysyms in buffer this request is immediately done */
  if (Keybuf.kb_count>0) { 
    unsigned keysym = Keybuf.kb_buf[Keybuf.kb_rkout];
    kbd_rqid = 0;
    Keybuf.kb_rkout = (Keybuf.kb_rkout+1) % KEYSYM_BUFLEN;
    Keybuf.kb_count--;
    (void) request_end(rqid, keysym);
  }
 end:
  int_restore(intlev);
  return rqid;
}

const char keymap_ident[] = KEYMAP_VERSION
 " <" __DATE__ "@" __TIME__ ">";

/**** FIXME: the keyboard irq handler is never called, why?? */
void init_keyboard(void) {
  /* flush keyboard */
  while(kb_read() != -1);
  /* self test - 0x55 means ok */
  kb_write(KB_CMD_REG, KBC_SELF_TEST);
  if(kb_read() != 0x55) 
    panic("keyboard self test failed.");
  /* kb interface test - 0 is ok */
  kb_write(KB_CMD_REG, KBC_IF_TEST);
  if(kb_read() != 0x00) 
    panic("keyboard interface test failed.");
  /* enable interface */
  kb_write(KB_CMD_REG, KBC_IF_ENABLE);	
  /* enable keyboard */
  kb_write(KB_DATA_REG, KB_ENABLE);
  if(kb_read() != KB_ACK) 
    panic("no ack to keyboard enable.");
  create_keybuf();
#ifdef KUT_DEBUG
  scr_c_show(" kbdirq ... ");
#endif /*KUT_DEBUG*/
  enable_irq(KB_IRQ);
#ifdef KUT_DEBUG
  scrs_sn("kbmap= " KEYMAP_VERSION);
  scrs_sn("[kutkeyb.c $Revision: 0.8 $] keyboard initialized!");
#endif /*KUT_DEBUG*/
} /* end of init_keyboard */


/*** rcslog ***
 * $Log: kutkeyb.c,v $
 * Revision 0.8  1998/09/02 20:29:08  basile
 * I still don't get keyboard interrupts
 *
 * Revision 0.7  1998/08/30 09:20:24  basile
 * int handler gets a stack (used with struct frame_st)
 *
 * Revision 0.6  1998/08/24 20:15:58  basile
 * keyboard irq handler never called... Either keyboard is disabled or I
 * have other tricks!
 *
 * Revision 0.5  1998/08/23 11:07:08  basile
 * why no keyboard interrupts?
 *
 * Revision 0.4  1998/08/22 21:20:48  basile
 * added debug messages
 *
 * Revision 0.3  1998/08/22 19:47:23  basile
 * added keyscan->keycode from Linux 2.1
 *
 *** endlog ***/

/** eof $Header: /home/basile/KUT/RCS/kutkeyb.c,v 0.8 1998/09/02 20:29:08 basile Exp $ **/

