/* File KUT/screen.c */
/* Time-stamp: <1998 Augu 23 Sun 11:47:11 CEST Basile.Starynkevitch@wanadoo.fr> */
/* rcsid: $Id: kutscreen.c,v 0.2 1998/08/23 11:10:02 basile Exp $ */

#include "kut.h"



const char screen_ident[] = 
"@(#)$Id: kutscreen.c,v 0.2 1998/08/23 11:10:02 basile Exp $"
 " <" __DATE__ "@" __TIME__ ">";


/* the hardware CGA screen memory - MSbyte is attribute, LSbyte is char */



#define SCRCELL(Attr,Chr) (((unsigned) ((unsigned char)(Attr))<<8) | ((unsigned char)(Chr)))


struct screen_struct_st {
  /* screen physical memory */
  unsigned short*scr_mem;
  /* read-only screen dimensions */
  unsigned short scr_width;
  unsigned short scr_height;
  /* read-only screen ports */
  unsigned short scr_port_idx;
  unsigned short scr_port_dat;
  /* current attribute */
  unsigned char scr_attr;
  /* current position */
  unsigned short scr_x, scr_y;
} Scr = {
  scr_width: 80,
  scr_height: 25,
  /* CGA ports */
  scr_port_idx: 0x3d4,
  scr_port_dat: 0x3d5,
  /* CGA memory */
  scr_mem: (unsigned short*)  0xB8000,
  scr_x: 0,
  scr_y: 0,
  scr_attr: SCR_ATTR_NORM
};


/* update the current hardware cursor */
void scr_update_cursor(void)
{
  unsigned long pos = (Scr.scr_y * Scr.scr_width + Scr.scr_x) << 1;
  outportb(Scr.scr_port_idx, 0xE);
  outportb(Scr.scr_port_dat, (pos >> 8) & 0xFF);
  outportb(Scr.scr_port_idx, 0xF);
  outportb(Scr.scr_port_dat, pos & 0xFF);
} /* end of scr_update_cursor */



/* clear the screen */
void scr_clear()
{
  int i;
  int d = Scr.scr_width * Scr.scr_height;  
  for (i=0; i<d; i++)
    Scr.scr_mem[i] = SCRCELL(SCR_ATTR_NORM,' ');
  Scr.scr_x = Scr.scr_y = 0;
  Scr.scr_attr = SCR_ATTR_NORM;
} /* end of scr_clear */

/* show a single character */
void scr_char_show(char c)
{
  unsigned short *scrbuf;
  unsigned  p;
  p = (Scr.scr_width * Scr.scr_y) + Scr.scr_x;
  scrbuf = Scr.scr_mem + p;
  *scrbuf = SCRCELL(Scr.scr_attr, ((unsigned char)c));
  if (Scr.scr_x<Scr.scr_width-1) Scr.scr_x++;
} /* end of scr_char_show */

/* show a string on a screen - no characters interpretation; don't
   wrap at end of line! */
void scr_raw_show(char *s, int len)
{
  unsigned short *scrbuf;
  unsigned p;
  p = (Scr.scr_width * Scr.scr_y) + Scr.scr_x;
  scrbuf = Scr.scr_mem + p;
  while (len>0 && Scr.scr_x < Scr.scr_width) {
    *scrbuf = SCRCELL(Scr.scr_attr, (unsigned char)(*s));
    Scr.scr_x++;
    scrbuf++, s++, len--;
  }
} /* end of scr_raw_show */

/* show a signed integer on the screen in decimal */
void scr_show_decint(int i) 
{
  /* stack allocated stuff should be really small! */
  char buf[12];
  char neg=0;
  unsigned short *scrbuf;
  unsigned p;
  int rank = 0;
  buf[0]=buf[1]=buf[2]=buf[3]=buf[4]=buf[5]=buf[6]=buf[7]=buf[8]='&';
  p = (Scr.scr_width * Scr.scr_y) + Scr.scr_x;
  scrbuf = Scr.scr_mem + p;
  /* remember sign */
  if (i<0) {
    neg='-';
    i=(-i);
  };
  /* special case for 0*/
  if (i==0) {
    if (Scr.scr_x+1 >= Scr.scr_width)
      return;
    *scrbuf = SCRCELL(Scr.scr_attr, '0');
    Scr.scr_x++;
    return;
  };
  while (i>0) {
    buf[rank++]='0'+(i%10);
    i /= 10;
  };
  if (Scr.scr_x+rank+1>=Scr.scr_width)
    return;
  if (neg) {
    *(scrbuf++) = SCRCELL(Scr.scr_attr, neg);
    Scr.scr_x ++;
  };
  while (rank>0) {
    rank--;
    *(scrbuf++) = SCRCELL(Scr.scr_attr, buf[rank]);
    Scr.scr_x ++;
  };
} /* end of scr_show_decint */



/* show an unsigned word on the screen in hexadecimal */
void scr_show_hexword(unsigned w) 
{
  unsigned short *scrbuf;
  unsigned p;
  char c;
  p = (Scr.scr_width * Scr.scr_y) + Scr.scr_x;
  scrbuf = Scr.scr_mem + p;
  if (Scr.scr_x + 8 >= Scr.scr_width) 
    return;
#define HANDLE_HEXDIGIT {                       \
  c = w & 0xf;					\
  w >>= 4;					\
  if (c<10)					\
    c += '0';					\
  else						\
    c += ('A' - 10);				\
  }
  HANDLE_HEXDIGIT;
  scrbuf[7] = SCRCELL(Scr.scr_attr,c);
  HANDLE_HEXDIGIT;
  scrbuf[6] = SCRCELL(Scr.scr_attr,c);
  HANDLE_HEXDIGIT;
  scrbuf[5] = SCRCELL(Scr.scr_attr,c);
  HANDLE_HEXDIGIT;
  scrbuf[4] = SCRCELL(Scr.scr_attr,c);
  HANDLE_HEXDIGIT;
  scrbuf[3] = SCRCELL(Scr.scr_attr,c);
  HANDLE_HEXDIGIT;
  scrbuf[2] = SCRCELL(Scr.scr_attr,c);
  HANDLE_HEXDIGIT;
  scrbuf[1] = SCRCELL(Scr.scr_attr,c);
  HANDLE_HEXDIGIT;
  scrbuf[0] = SCRCELL(Scr.scr_attr,c);
  Scr.scr_x += 8;
}

/* show a C-null-terminated-string on a screen - no characters
   interpretation; don't wrap at end of line! */
void scr_c_show(char *s)
{
  unsigned short *scrbuf;
  unsigned p;
  char c;
  p = (Scr.scr_width * Scr.scr_y) + Scr.scr_x;
  scrbuf = Scr.scr_mem + p;
  while (Scr.scr_x < Scr.scr_width && *s) {
    c = *s;
    *scrbuf = SCRCELL(Scr.scr_attr, (unsigned char)c);
    Scr.scr_x++;
    scrbuf++, s++;
  }
  scr_update_cursor();
} /* end of scr_c_show */

/* new line on screen, with scrolling if needed */
void scr_nl(void)
{
  unsigned short *scrbuf;
  int i,j;
  Scr.scr_x = 0;
  if (Scr.scr_y < Scr.scr_height-1) {
    Scr.scr_y ++;
  } else {
    Scr.scr_y = Scr.scr_height-1;
    /* scroll one line */
    for (i=0; i<Scr.scr_y; i++) {
      scrbuf = Scr.scr_mem + (Scr.scr_width * i);
      for (j=Scr.scr_width; j>0; j--) {
	scrbuf[0] = scrbuf[Scr.scr_width];
	scrbuf++;
      }
    };
    /* clear last line */
    scrbuf = Scr.scr_mem + (Scr.scr_y * Scr.scr_width);
    for (i=Scr.scr_width; i>0; i--, scrbuf++) 
      *scrbuf =  SCRCELL(Scr.scr_attr, (unsigned char)' ');
  }
  scr_update_cursor();
} /* end of scr_nl */

void scr_maybenl(void)
{
  if (Scr.scr_x>0)
    scr_nl();
}
/* set cursor position */
void scr_setpos(int x, int y)
{
  if (x<0)
    x = Scr.scr_x;
  if (y<0)
    y = Scr.scr_y;
  if (x>=Scr.scr_width) x = Scr.scr_width-1;
  if (y>=Scr.scr_height) y = Scr.scr_height-1;
  Scr.scr_x = x;
  Scr.scr_y = y;
  scr_update_cursor();
} /* end of scr_setpos */

void scr_setattr(int a)
{
  Scr.scr_attr = a;
}

int scr_posx(void)
{
  return Scr.scr_x;
}

int scr_posy(void)
{
  return Scr.scr_y;
}
/* eof $Header: /home/basile/KUT/RCS/kutscreen.c,v 0.2 1998/08/23 11:10:02 basile Exp $ */
