(* file KUT/ut_kbdmap.ml -*- caml -*- *)
(* Time-stamp: <1998 Augu 22 Sat 09:34:42 CEST Basile.Starynkevitch@wanadoo.fr> *)

(* rcsid: $Id: ut_kbdmap.ml,v 0.7 1998/08/22 19:48:10 basile Exp $ *)
(* some stuff from Linux kbd-0.96 ksym.c file *)

(* In Unicode and ISO10646-1 the range 0xe000 to 0xf8ff will never be
assigned any characters by the standard and is reserved for private
usage. 

   Meta+Ascii N are coded 0xf800+N
   function keys F1, F2... starting from 0xf900
   arrows starting from 0xfa00
*)

type keysym_t =
    Nokey
  | Iso of char
  | Kmeta of char
  | Arrow of arrow_t
  | Modifier of modifier_t
  | Dead of dead_t
  | Function of int
  | Special of special_t
  | Keypad of keypad_t
  | Ascii of int (* ascii digit *)

and
    modifier_t = 
    Nomod 
    | Shift 
    | AltGr 
    | Control
    | Alt 
    | CapsShift 
    | Compose
    | Numlock
    | Meta
    | Scrolllock

and dead_t = Deadnone 
 | Dead_grave 
 | Dead_acute 
 | Dead_circumflex
 | Dead_tilde 
 | Dead_diaeresis 
 | Dead_cedilla

and arrow_t = Noarrow
 | Down
 | Left
 | Right
 | Up
 | Home
 | End
 | Prior
 | Next
 | Scrollbackward
 | Scrollforward

and keypad_t = Kpnone
 | KP_0 
 | KP_1 
 | KP_2 
 | KP_3
 | KP_4 
 | KP_5 
 | KP_6 
 | KP_7 
 | KP_8 
 | KP_9 
 | KP_Add
 | KP_Subtract 
 | KP_Multiply
 | KP_Divide
 | KP_Enter
 | KP_Comma
 | KP_Period
 | KP_MinPlus

and special_t = Spnone
 | Return
 | Bigreturn
 | Break
 | Boot
 | Debug

;;

let keysym_dict = Hashtbl.create 311 ;;

let error_cnt = ref 0;;

let errprintf fmt = incr error_cnt; Printf.eprintf fmt;;

let escaped_char c = 
  if (c=='\'') then "\\'" 
  else if (Char.uppercase c != Char.lowercase c) then String.make 1 c 
  else 
    let sc = String.escaped (String.make 1 c)
    in if ((String.length sc)>2) then Printf.sprintf "\\%03o" (Char.code c) else sc;;
  
let isolatin1_symlist = [ (*list of ISOLatin1 symbols*) 

(* 0= *) "nul"; "Control_a"; "Control_b"; "Control_c"; "Control_d";
"Control_e"; "Control_f"; "Control_g"; "BackSpace"; "Tab";

(* 10= *) "Linefeed"; "Control_k"; "Control_l"; "Control_m";
"Control_n"; "Control_o"; "Control_p"; "Control_q"; "Control_r";
"Control_s";

(* 20= *) "Control_t"; "Control_u"; "Control_v"; "Control_w";
"Control_x"; "Control_y"; "Control_z"; "Escape";
"Control_backslash"; "Control_bracketright";

(* 30= *) "Control_asciicircum"; "Control_underscore"; "space";
"exclam"; "quotedbl"; "numbersign"; "dollar"; "percent";
"ampersand"; "apostrophe";

(* 40= *) "parenleft"; "parenright"; "asterisk"; "plus"; "comma";
"minus"; "period"; "slash"; "zero"; "one";

(* 50=*) "two"; "three"; "four"; "five"; "six"; "seven"; "eight"; 
"nine"; "colon"; "semicolon";

(* 60= *) "less"; "equal"; "greater"; "question"; "at"; "A"; "B";
 "C"; "D"; "E";

(* 70= *) "F"; "G"; "H"; "I"; "J"; "K"; "L"; "M"; "N"; "O";

 (* 80= *) "P"; "Q"; "R"; "S"; "T"; "U"; "V"; "W"; "X"; "Y";

(* 90=*) "Z"; "bracketleft"; "backslash"; "bracketright";
"asciicircum"; "underscore"; "grave"; "a"; "b"; "c";

(* 100=*) "d"; "e"; "f"; "g"; "h"; "i"; "j"; "k"; "l"; "m";

(* 110=*) "n"; "o"; "p"; "q"; "r"; "s"; "t"; "u"; "v"; "w";

(*  120= *) "x"; "y"; "z"; "braceleft"; "bar"; "braceright";
"asciitilde"; "Delete"; ""; "";

(* 130=*) ""; ""; ""; ""; ""; ""; ""; ""; ""; "";

(* 140=*) ""; ""; ""; ""; ""; ""; ""; ""; ""; "";

(* 150=*) ""; ""; ""; ""; ""; ""; ""; ""; ""; "";

(* 160=*) "nobreakspace"; "exclamdown"; "cent"; "sterling";
"currency"; "yen"; "brokenbar"; "section"; "diaeresis";
"copyright";

(* 170=*) "ordfeminine"; "guillemotleft"; "notsign"; "hyphen";
"registered"; "macron"; "degree"; "plusminus"; "twosuperior";
"threesuperior";

(* 180=*) "acute"; "mu"; "paragraph"; "periodcentered"; "cedilla"
; "onesuperior"; "masculine"; "guillemotright"; "onequarter";
"onehalf";

(* 190=*) "threequarters"; "questiondown"; "Agrave"; "Aacute";
"Acircumflex"; "Atilde"; "Adiaeresis"; "Aring"; "AE"; "Ccedilla";

(* 200= *) "Egrave"; "Eacute"; "Ecircumflex"; "Ediaeresis";
"Igrave"; "Iacute"; "Icircumflex"; "Idiaeresis"; "ETH"; "Ntilde";

(* 210= *) "Ograve"; "Oacute"; "Ocircumflex"; "Otilde";
"Odiaeresis"; "multiply"; "Ooblique"; "Ugrave"; "Uacute";
"Ucircumflex";

(* 220= *) "Udiaeresis"; "Yacute"; "THORN"; "ssharp"; "agrave";
"aacute"; "acircumflex"; "atilde"; "adiaeresis"; "aring";

(* 230= *) "ae"; "ccedilla"; "egrave"; "eacute"; "ecircumflex";
"ediaeresis"; "igrave"; "iacute"; "icircumflex"; "idiaeresis";

(* 240= *) "eth"; "ntilde"; "ograve"; "oacute"; "ocircumflex";
"otilde"; "odiaeresis"; "division"; "oslash"; "ugrave";

(* 250= *) "uacute"; "ucircumflex"; "udiaeresis"; "yacute";
"thorn"; "ydiaeresis" 
] ;;


let rec add_isolatin1 rank = 
  function nm :: rest 
      -> ( if (nm != "") then Hashtbl.add keysym_dict nm (Iso (Char.chr rank)) );
    	add_isolatin1 (rank+1) rest
    |  [] -> ()
in add_isolatin1 0 isolatin1_symlist ;;

let rec add_meta rank = 
  function nm :: rest 
      -> ( if (nm != "") then Hashtbl.add keysym_dict ("Meta_" ^ nm) (Kmeta (Char.chr rank)) );
    	add_meta (rank+1) rest
    |  [] -> ()
in add_meta 0 isolatin1_symlist ;;


List.iter (fun (n,a) -> Hashtbl.add keysym_dict n (Arrow a)) 
[ "Down", Down; "Left", Left; "Right", Right; "Up", Up; 
  "End", End; "Home", Home; "Prior", Prior; "Next", Next;
  "Scrollbackward", Scrollbackward; "Scroll_Backward", Scrollbackward (*variant*);
  "Scrollforward", Scrollforward; "Scroll_Forward", Scrollbackward (*variant*) ];;

let arrow_name = function 
    Down -> "Down" |  Left -> "Left" | Right -> "Right" | Up -> "Up"
  | End -> "End" | Home -> "Home" | Prior -> "Prior" | Next -> "Next"
  | Scrollbackward -> "Scrollbackward" | Scrollforward -> "Scrollforward"
  |  _ -> failwith "arrow_name" ;;

let out_arrow chan ar = output_string chan (arrow_name ar);;

List.iter (fun (n,m) -> Hashtbl.add keysym_dict n (Modifier m)) 
[ "Shift", Shift ; "AltGr", AltGr ; "Control", Control ; "Alt", Alt ;
  "CapsShift", CapsShift  ; "Caps_Lock", CapsShift (*variant*) ;
  "Meta", Meta; "Compose", Compose; 
  "Numlock", Numlock;  "Num_Lock", Numlock (*variant*);
  "Scrolllock", Scrolllock; "Scroll_Lock", Scrolllock (*variant*) ];;

let modifier_name = function
    Shift-> "Shift" | AltGr-> "AltGr" | Control-> "Control" | Alt-> "Alt" 
  | CapsShift-> "CapsShift" | Meta -> "Meta" 
  | Compose-> "Compose" | Numlock -> "Numlock"
  | Scrolllock -> "Scrolllock"
  |  _ -> failwith "modifier_name";;

let out_modifier chan md = output_string chan (modifier_name md);;

List.iter (fun (n,d) -> Hashtbl.add keysym_dict n (Dead d))
[ "dead_grave", Dead_grave  ; "dead_acute", Dead_acute;
"dead_circumflex", Dead_circumflex; "dead_tilde", Dead_tilde; 
"dead_diaeresis", Dead_diaeresis; "dead_cedilla",  Dead_cedilla] ;;

let dead_name = function
   Dead_grave -> "dead_grave"
   | Dead_acute -> "dead_acute"
   | Dead_circumflex -> "dead_circumflex"
   | Dead_tilde -> "dead_tilde"
   | Dead_diaeresis -> "dead_diaeresis"
   | Dead_cedilla ->  "dead_cedilla" 
   |  _ -> failwith "dead_name";;

let dead_equivalent = function
   Dead_grave -> '`'
   | Dead_acute -> '\''
   | Dead_circumflex -> '^'
   | Dead_tilde -> '~'
   | Dead_diaeresis -> ''
   | Dead_cedilla ->  ','
   |  _ -> failwith "dead_equivalent";;

let dead_list = [ Dead_grave; Dead_acute; Dead_circumflex; Dead_tilde;
Dead_diaeresis; Dead_cedilla ];;


let char_dead = function
    '`'  ->  Dead_grave     
   |'\'' ->  Dead_acute     
   |'^'  ->  Dead_circumflex
   |'~'  ->  Dead_tilde     
   |''  ->  Dead_diaeresis 
   | ',' ->  Dead_cedilla   
   | _ -> Deadnone;;

let out_dead chan d = output_string chan (dead_name d);;

let keypad_name = function                                        
KP_0 ->	   "KP_0"              
 | KP_1 ->      "KP_1"             
 | KP_2 ->      "KP_2"             
 | KP_3 ->      "KP_3"             
 | KP_4 ->      "KP_4"             
 | KP_5 ->      "KP_5"             
 | KP_6 ->      "KP_6"             
 | KP_7 ->      "KP_7"             
 | KP_8 ->      "KP_8"             
 | KP_9 ->      "KP_9"       
 | KP_Add	->    "KP_Add"           
 | KP_Subtract 	->    "KP_Subtract"      
 | KP_Multiply	->    "KP_Multiply"      
 | KP_Divide	->    "KP_Divide"        
 | KP_Enter	->    "KP_Enter"         
 | KP_Comma	->    "KP_Comma"         
 | KP_Period	->    "KP_Period"        
 | KP_MinPlus 	->    "KP_MinPlus"        
 |  _ -> failwith "keypad_name"
;;

let out_keypad chan kp = output_string chan (keypad_name kp);;
List.iter (fun (n,k) -> Hashtbl.add keysym_dict n (Keypad k)) [
"KP_0", KP_0; "KP_1", KP_1; "KP_2", KP_2; "KP_3", KP_3; "KP_4", KP_4;
"KP_5", KP_5; "KP_6", KP_6; "KP_7", KP_7; "KP_8", KP_8; "KP_9", KP_9;
"KP_Add", KP_Add; "KP_Subtract", KP_Subtract; 
"KP_Multiply", KP_Multiply; "KP_Divide", KP_Divide; 
"KP_Enter", KP_Enter; "KP_Comma", KP_Comma; "KP_Period", KP_Period;
"KP_MinPlus", KP_MinPlus ] ;;

let special_name = function
   Return      -> "Return"      
 | Bigreturn   -> "Bigreturn"   
 | Break       -> "Break"       
 | Boot        -> "Boot"        
 | Debug       -> "Debug"
 |  _ -> failwith "special_name"
;;
            
let out_special chan sp = output_string chan (special_name sp);;

List.iter (fun (n,s) -> Hashtbl.add keysym_dict n (Special s)) [
"Return"       ,  Return      ;	    
"Bigreturn"    ,  Bigreturn   ;	   
"Break"        ,  Break       ;	   
"Boot"         ,  Boot        ;	   
"Debug"        ,  Debug       
] ;;	    




let negfunc_symlist = [ (* negative coded function *) "" ; "Find";
"Insert"; "Remove"; "Select"; "Macro"; "Help"; "Do";
"Pause" ];;




let rec add_negfunc rank = 
  function nm :: rest 
      -> ( if (nm != "") then Hashtbl.add keysym_dict nm (Function (-rank)) );
    	add_negfunc (rank+1) rest
    |  [] -> ()
in add_negfunc 0 negfunc_symlist ;;


for fc = 0 to 100 do 
  let fnam = "F" ^ (string_of_int fc)
  in
  Hashtbl.add keysym_dict fnam (Function fc)
done;;

for fc = 0 to 9 do 
  let fnam = "Ascii_" ^ (string_of_int fc)
  in
  Hashtbl.add keysym_dict fnam (Ascii fc)
done;;

Hashtbl.add keysym_dict "Ascii_none" (Ascii (-1));;

let keymap_version = ref "" ;;

let keymap_matrix = Array.create_matrix 256 16 Nokey ;;

let compose_matrix = Array.create_matrix 256 256 None ;;

let composed_flag = Array.create 256 false;;

let keymap_count = ref 0 ;;

let rec 
    fill_keymap code rank = function
	[] -> ()
      |	ks::rest 
	->  keymap_matrix.(code).(rank) <- ks; 
	  incr keymap_count;
	  fill_keymap code (rank+1) rest

and
    parse_keymap_entry = parser
      	[< 'Genlex.Kwd "keycode" ; 'Genlex.Int code ; 
	  'Genlex.Kwd "=" ; ksl = parse_keysyms_list >]
 	-> fill_keymap code 0 ksl
  |  [< 'Genlex.Kwd "alt" ; 'Genlex.Kwd "keycode" ; 'Genlex.Int code ;
       'Genlex.Kwd "=" ; ksl = parse_keysyms_list >]
    ->  fill_keymap code 8 ksl
  |  [< 'Genlex.Kwd "altgr" ; 'Genlex.Kwd "keycode" ; 'Genlex.Int code ;
       'Genlex.Kwd "=" ; ksl = parse_keysyms_list >]
    ->  fill_keymap code 2 ksl
  |  [< 'Genlex.Kwd "control" ; 'Genlex.Kwd "keycode" ; 'Genlex.Int code ;
       'Genlex.Kwd "=" ; ksl = parse_keysyms_list >]
    ->  fill_keymap code 4 ksl
  |  [< 'Genlex.Kwd "ctrlalt" ;  'Genlex.Kwd "keycode" ; 'Genlex.Int code ;
       'Genlex.Kwd "=" ; ksl = parse_keysyms_list >]
    ->  fill_keymap code 6 ksl
  | [< 'Genlex.Kwd "version" ;  'Genlex.String vers >]
    ->  keymap_version := vers
  | [< 'Genlex.Kwd "message" ;  'Genlex.String msg >]
    ->  
      Printf.eprintf "message %s\n" msg; flush stderr
  | [<  'Genlex.Kwd "compose" ;  'Genlex.Char c1;  'Genlex.Char c2; 
      'Genlex.Kwd "to" ;  'Genlex.Char cc >]
    -> 
      compose_matrix.(Char.code c1).(Char.code c2) <- Some cc;
      compose_matrix.(Char.code c2).(Char.code c1) <- Some cc;
      composed_flag.(Char.code c1) <- true;
      composed_flag.(Char.code c2) <- true
and
    parse_keysyms_list = parser 
	[< 'Genlex.Ident name ; rest = parse_keysyms_list >] -> 
	  begin
	    try 
	      (Hashtbl.find keysym_dict name) :: rest
	    with
	      Not_found -> 
		errprintf "unknown keysym name=%s\n" name; flush stderr;
		rest
	  end
      |	[< 'Genlex.Int code ; rest = parse_keysyms_list >] -> 
	  errprintf "ignored numerical keysym %d\n" code; flush stderr;
	  rest
      | [< >] -> []

;;

let main_prog () =
  prerr_endline "running _utkbdmap";
  let keymap_name = Sys.argv.(1) and cprog_name = Sys.argv.(2) in
  Printf.eprintf " begin keymap_name=%s cprog_name=%s" keymap_name cprog_name;
  prerr_newline ();
  if not (Sys.file_exists keymap_name) then begin
    prerr_endline ("no such keymap file: " ^ keymap_name);
    exit 1
  end;
  begin
    let keymap_in = open_in keymap_name and cprog_out = open_out cprog_name in
    let cprintf fmt = (Printf.fprintf cprog_out fmt)
    and cputs = output_string cprog_out
    in
    cprintf "/** DO NOT EDIT -*- C -*- file %s generated from %s **/\n" 
      cprog_name keymap_name  ;
    cputs "/**... by $Id: ut_kbdmap.ml,v 0.7 1998/08/22 19:48:10 basile Exp $ **/\n";
    let keymap_lexer = 
      Genlex.make_lexer [ "=" ; "+" ; 
			  "version" ; "message" ; "compose" ; "to" ;
			  "keycode" ; "plain" ; "control" ; "alt" ; "altgr" ;
			  "ctrlalt"  ] 
	(Stream.of_channel keymap_in)
	in
    flush stderr;
    flush cprog_out;
    let rec parser_all = 
      parser
	[< ks = parse_keymap_entry ; rest >] -> parser_all rest
	  | [< >] -> ()
     in
    parser_all keymap_lexer ;
    Printf.eprintf "parsed %d key entries\n" (!keymap_count); flush stderr;
    if (!keymap_version != "") then 
      cprintf "#define KEYMAP_VERSION \"%s\"\n" (String.escaped !keymap_version);
    cputs "#define LOGSHIFT    0\n";
    cputs "#define LOGALTGR    1\n";
    cputs "#define LOGCONTROL  2\n";
    cputs "#define LOGALT      3\n";
    cputs "#define LOGCAPS     4\n";
    cputs "#define LOGCOMPOSE  5\n";
    cputs "#define LOGMETA     6\n";
    cputs "#define LOGNUM      7\n";
    cputs "#define LOGSCROLL   8\n";

    cputs "#define MODSHIFT    (1<<LOGSHIFT)\n";
    cputs "#define MODALTGR    (1<<LOGALTGR)\n";
    cputs "#define MODCONTROL  (1<<LOGCONTROL)\n";
    cputs "#define MODALT      (1<<LOGALT)\n";
    cputs "#define MODCAPS     (1<<LOGCAPS)\n";
    cputs "#define MODCOMPOSE  (1<<LOGCOMPOSE)\n";
    cputs "#define MODMETA     (1<<LOGMETA)\n";
    cputs "#define MODNUM      (1<<LOGNUM)\n";
    cputs "#define MODSCROLL   (1<<LOGSCROLL)\n";
    cputs "/* 0xe000 to 0xf8ff is not assigned in Unicode */\n";
    cputs "#define CTRLOF(C) ((C)&0x1f)\n";
    cputs "#define METAOF(C) ((C)|0xf800) /*0xf8YY unassigned in Unicode*/\n";
    cputs "#define FUNCTIONOF(N) ((N)|0xf900)\n";
    cputs "#define KS_BREAK          0xf9ff\n";
    cputs "/*arrow starting from 0xfa00 */\n";
    cputs "#define KS_ARRUP          0xfa00\n";
    cputs "#define KS_ARRDOWN        0xfa01\n";
    cputs "#define KS_ARRLEFT        0xfa02\n";
    cputs "#define KS_ARRRIGHT       0xfa03\n";
    cputs "#define KS_ARRPRIOR       0xfa04\n";
    cputs "#define KS_ARRNEXT        0xfa05\n";
    cputs "#define KS_ARRHOME        0xfa06\n";
    cputs "#define KS_ARREND         0xfa07\n";
    cputs "#define KS_SCROLLBACK     0xfa10\n";
    cputs "#define KS_SCROLLFORWARD  0xfa11\n";
    cputs "#define KS_SCROLLLOCK     0xfa20\n";
    cputs "#define KS_SCROLLUNLOCK   0xfa21\n";
    cputs "/*keypad starting from 0xfa80 */\n";
    cputs "#define KS_KP_0           0xfa80\n" ;  
    cputs "#define KS_KP_1 	     0xfa81\n" ;  
    cputs "#define KS_KP_2 	     0xfa82\n" ;  
    cputs "#define KS_KP_3	     0xfa83\n" ;  
    cputs "#define KS_KP_4 	     0xfa84\n" ;  
    cputs "#define KS_KP_5 	     0xfa85\n" ;  
    cputs "#define KS_KP_6 	     0xfa86\n" ;  
    cputs "#define KS_KP_7 	     0xfa87\n" ;  
    cputs "#define KS_KP_8 	     0xfa88\n" ;  
    cputs "#define KS_KP_9 	     0xfa89\n" ;  
    cputs "#define KS_KP_Add	     0xfa8a\n" ;  
    cputs "#define KS_KP_Subtract    0xfa8b\n" ;  
    cputs "#define KS_KP_Multiply    0xfa8c\n" ;  
    cputs "#define KS_KP_Divide	     0xfa8d\n" ;  
    cputs "#define KS_KP_Enter	     0xfa8e\n" ;  
    cputs "#define KS_KP_Comma	     0xfa8f\n" ;  
    cputs "#define KS_KP_Period	     0xfa90\n" ;  
    cputs "#define KS_KP_MinPlus     0xfa91\n" ;	

    cputs "static void\n";
    cputs "dispatch_keycode(unsigned code, unsigned pressed) {\n";
    cputs "  static int modified=0;\n";
    cputs "  static int toggled=0;\n";
    cputs "  static int composed_symnum= -1;\n";
    cputs "  static int dead_accent= 0;\n";
    cputs "  int symnum= -1;\n";
    cputs "#define EMIT_SYMNUM(N) {symnum=(N); goto emit_symnum;}\n";
    cputs "#define EMIT_COMPSYM() {symnum=composed_symnum; composed_symnum=(-1); goto emit_symnum;}\n";
    cputs "  if (!modified)\n";
    cputs "    switch(code) /* test if modifier */ {\n";
    for cod = 0 to 255 do
      match keymap_matrix.(cod).(0) with
	Modifier Shift -> 
	  cprintf "    case %d /* Shift */:\n" cod;
	  cputs "     if (pressed) modified |= MODSHIFT; else modified &= (0xffff^MODSHIFT);\n";
	  cputs "     return;\n"
      |	Modifier AltGr ->
	  cprintf "    case %d /* AltGr */:\n" cod;
	  cputs "     if (pressed) modified |= MODALTGR; else modified &= (0xffff^MODALTGR);\n";
	  cputs "     return;\n"
      |	Modifier Control ->
	  cprintf "    case %d /* Control */:\n" cod;
	  cputs "     if (pressed) modified |= MODCONTROL; else modified &= (0xffff^MODCONTROL);\n";
	  cputs "     return;\n"
      |	Modifier Alt  ->
	  cprintf "    case %d /* Alt */:\n" cod;
	  cputs "     if (pressed) modified |= MODALT; else modified &= (0xffff^MODALT);\n";
	  cputs "     return;\n"
      |	Modifier CapsShift ->
	  cprintf "    case %d /* CapsShift */:\n" cod;
	  cputs "     if (pressed) toggled ^= MODCAPS; /* release Capshift is ignored */\n";
	  cputs "     show_caps_led(toggled&MODCAPS);\n";
	  cputs "     return;\n"
      |	Modifier Compose ->
	  cprintf "    case %d /* Compose */:\n" cod;
	  cputs "     if (pressed) modified |= MODCOMPOSE;\n";
	  cputs "     else {\n";
	  cputs "        modified &= (0xffff^MODCOMPOSE);\n";
	  cputs "        if (composed_symnum >= 0) EMIT_COMPSYM();\n";
	  cputs "     };\n";
	  cputs "     composed_symnum= (-1);\n";
	  cputs "     return;\n"
      |	Modifier Numlock ->
	  cprintf "    case %d /* Numlock */:\n" cod;
	  cputs "     if (pressed) toggled ^= MODNUM; /* release Numlock is ignored */\n";
	  cputs "     show_nums_led(toggled&MODNUM);\n";
	  cputs "     return;\n"
      |	Modifier Meta  ->
	  cprintf "    case %d /* Meta */:\n" cod;
	  cputs "     if (pressed) modified |= MODMETA; else modified &= (0xffff^MODMETA);\n";
	  cputs "     return;\n"
      |	Modifier Scrolllock ->
	  cprintf "    case %d /* Scrolllock */:\n" cod;
	  cputs "     if (pressed) toggled ^= MODSCROLL; /* release Scrolllock is ignored */\n";
	  cputs "     show_scroll_led(toggled&MODSCROLL);\n";
	  cputs "     if (toggled&MODSCROLL) {\n";
	  cputs "       EMIT_SYMNUM(KS_SCROLLLOCK);\n";
	  cputs "     } else {\n";
	  cputs "       EMIT_SYMNUM(KS_SCROLLUNLOCK);\n";
	  cputs "     };\n";
	  cputs "     return;\n"
      |	Modifier _ ->
	  cprintf "    case %d /* unexpected modifier */:\n" cod ;
	  cputs "     return;\n";
      |	 _ -> ()
    done;
    cputs "    default:;\n";
    cputs "    } /* end switch for modifiers */\n";
    flush cprog_out;
    cputs "    if (!pressed) switch (code) /* main dispatch */ {\n";
    for cod = 0 to 255 do
      let keymap_row = keymap_matrix.(cod) in
      let nonempty_rest = 
	let rec scan_row rk = 
	  if (keymap_row.(rk) != Nokey) then true
	  else if (rk<15) then scan_row (rk+1) else false
	in scan_row 1
      in
      let nonempty_row = (keymap_row.(0) != Nokey) or nonempty_rest 
      in
      if (nonempty_row) then 
	begin
	  cprintf "   case /*keycode*/ %d: \n" cod;
	  flush cprog_out;
	  if (nonempty_rest or (
	  (* this match is a condition - return true if non single letter *)
	    match keymap_row.(0) with
	      Iso c when not (Char.lowercase c = Char.uppercase c) -> 
		let uc = Char.uppercase c 
		and lc = Char.lowercase c
		in 
		let sc = if (c=lc) then uc else lc in
		let output_compose x = 
		  for cc = 0 to 255 do 
		    match compose_matrix.(Char.code x).(cc) with
		      None -> ()
		    | Some k -> 
			cprintf "          else if (composed_symnum=='%s') composed_symnum='%s';\n"
			  (escaped_char (Char.chr cc)) (escaped_char k)
		  done
		and output_dead x = 
		  let handle_dead d = 
		    let cd = dead_equivalent d in
		    match compose_matrix.(Char.code x).(Char.code cd) with
		      None -> ()
		    | Some k -> 
			cprintf "       if (dead_accent=='%s') EMIT_SYMNUM('%s');\n" 
			  (escaped_char cd) (escaped_char k)
		  in List.iter handle_dead dead_list
		in
		cprintf "    /* letter %c shifted %c */\n"  c sc;
		cputs   "      if (modified==0||((toggled&MODCAPS) && modified==MODSHIFT)) {\n";
		output_dead c;
		cprintf "         EMIT_SYMNUM('%c');\n" c;
		cputs   "      };\n";
		cputs   "      if ((modified==MODSHIFT && !(toggled&MODCAPS)) || (toggled&MODCAPS)) {\n";
		output_dead sc;
		cprintf "         EMIT_SYMNUM('%c');\n" sc;
		cputs   "      };\n";
		cprintf "      if (modified==MODMETA) EMIT_SYMNUM(METAOF('%c'));\n" c;
		cprintf "      if (modified==(MODSHIFT|MODMETA)) EMIT_SYMNUM(METAOF('%c'));\n" sc;
		cprintf "      if (modified&MODCONTROL) EMIT_SYMNUM(CTRLOF('%c'));\n" uc;
		if (composed_flag.(Char.code c)) then begin
		  cputs   "      if ((modified&MODCOMPOSE)\n";
		  cputs   "         && !((modified&MODSHIFT) ^ (toggled&MODCAPS))) { \n";
		  cprintf "          if (composed_symnum<0) composed_symnum='%c';\n" c;
		  output_compose c;
		  cprintf   "      } /*end if compose %c*/\n" c;
		end;
		if (composed_flag.(Char.code sc)) then begin
		  cputs   "      if ((modified&MODCOMPOSE)\n";
		  cputs   "          && ((modified&MODSHIFT) ^ (toggled&MODCAPS))) {\n";
		  cprintf "          if (composed_symnum<0) composed_symnum='%c';\n" sc;
		  output_compose sc;
		  cprintf   "      } /*end if compose %c*/\n" sc;
		end;
		(* the match returns *) false
              | _ -> true
	  )) then begin
	    (* keyrow is not a single letter *)
	    for i = 0 to 4 do
	      let modifier_condstr = 
		[| "((modified&(MODSHIFT|MODALTGR|MODCONTROL|MODALT))==0)" ; 
		  "(modified&MODSHIFT)" ;
		  "(modified&MODALTGR)"; 
		  "(modified&MODCONTROL)";
		  "(modified&MODALT)" |] 
		  .(i) in
	      match keymap_row.(i) with
		Nokey -> ()
	      |	Modifier _ -> () (* already generated *)
	      |	Iso c when (Char.uppercase c != c) -> 
		  cprintf "   if %s {\n" modifier_condstr;
		  cprintf   "    if (toggled&MODCAPS) EMIT_SYMNUM('%s') else EMIT_SYMNUM('%s');\n"  
		    (escaped_char (Char.uppercase c)) (escaped_char c);
		  cputs   "   };\n"
	      |	Iso c  -> 
		  cprintf "   if %s EMIT_SYMNUM('%s');\n" modifier_condstr (escaped_char c)
	      |	Kmeta c -> 
		  cprintf "  if %s EMIT_SYMNUM(METAOF('%s'));\n" modifier_condstr (escaped_char c)
	      |	Dead d -> 
		  cprintf "   if %s dead_accent='%s';\n" modifier_condstr (escaped_char (dead_equivalent d));
	      |	Function f when (f>0) -> 
		  cprintf "   if %s EMIT_SYMNUM(FUNCTIONOF(%d));\n" modifier_condstr f
	      |	Function f when (f<0)  -> 
		  cprintf "   if %s EMIT_SYMNUM(FUNCTIONOF(200 %d));\n" modifier_condstr f
	      |	Special Return ->
		  cprintf "   if %s { /* Return */\n" modifier_condstr;
		  cputs   "    EMIT_SYMNUM(/*return*/'\\r');\n";
		  cputs   "   } /* end of Return */\n"
	      |	Special Break ->
		  cprintf "   if %s { /* Break */\n" modifier_condstr;
		  cputs   "    EMIT_SYMNUM(/*break*/KS_BREAK);\n";
		  cputs   "   } /* end of break */\n"
	      |	Keypad k ->
		  cprintf "   if %s { /*keypad*/\n" modifier_condstr;
		  cprintf "    EMIT_SYMNUM(/*keypad*/ KS_%a);\n" out_keypad k;
		  cputs   "   } /* end of keypad */\n"
	      |	Arrow Down
		-> cprintf "   if %s /*arrow down*/ EMIT_SYMNUM(KS_ARRDOWN);\n" modifier_condstr;
	      |	Arrow Left
 		-> cprintf "   if %s /*arrow left*/ EMIT_SYMNUM(KS_ARRLEFT);\n" modifier_condstr;
	      |	Arrow Right
 		-> cprintf "   if %s /*arrow right*/ EMIT_SYMNUM(KS_ARRRIGHT);\n" modifier_condstr;
	      |	Arrow Up
 		-> cprintf "   if %s /*arrow up*/ EMIT_SYMNUM(KS_ARRUP);\n" modifier_condstr;
	      |	Arrow Home
 		-> cprintf "   if %s /*arrow home*/ EMIT_SYMNUM(KS_ARRHOME);\n" modifier_condstr;
	      |	Arrow End
 		-> cprintf "   if %s /*arrow end*/ EMIT_SYMNUM(KS_ARREND);\n" modifier_condstr;
	      |	Arrow Prior
 		-> cprintf "   if %s /*arrow prior*/ EMIT_SYMNUM(KS_ARRPRIOR);\n" modifier_condstr;
	      |	Arrow Next
 		-> cprintf "   if %s /*arrow next*/ EMIT_SYMNUM(KS_ARRNEXT);\n" modifier_condstr;
	      |	Arrow Scrollbackward
 		-> cprintf "   if %s /*arrow scrollback*/ EMIT_SYMNUM(KS_SCROLLBACK);\n" modifier_condstr;
	      |	Arrow Scrollforward
 		-> cprintf "   if %s /*arrow scrollforward*/ EMIT_SYMNUM(KS_SCROLLFORWARD);\n" modifier_condstr;
	      |	Arrow a -> 
		  cprintf "   if %s { /*arrow*/\n" modifier_condstr;
		  cprintf "#warning code for arrow %a\n" out_arrow a;
		  cputs   "   } /* end of arrow */\n"
	      |	 _ -> errprintf "can't handle keycode %d /%d\n" cod i
	    done
	  end;
	  cputs "   break;\n"
	end
    done;
    cputs "   } /* end switch main dispatch */\n";
    cputs "   return;\n";
    cputs "emit_symnum:\n";
    cputs "  dead_accent = 0;\n";
    cputs "  emit_keysym(symnum);\n";
    cputs "} /* end of dispatch_keycode */\n\n";
    if (!error_cnt>0) then
	cprintf "\n#error got %d errors from keymap %s\n" !error_cnt keymap_name;
    cprintf "/*- end of generated file %s -*/\n" cprog_name;
    flush cprog_out;
    close_out cprog_out
  end;
  Printf.eprintf "generated key dispatch file %s from map %s\n"
    cprog_name keymap_name;
  flush stderr;
  if (!error_cnt>0) then begin
    Printf.eprintf "** got %d errors from keymap %s\n" !error_cnt keymap_name; 
    flush stderr;
    exit 1
  end
;;
Printexc.catch main_prog ();;

(*** rcslog ****
 * $Log: ut_kbdmap.ml,v $
 * Revision 0.7  1998/08/22 19:48:10  basile
 * dispatch_keycode and emit_keysym don't need any data
 *
 * Revision 0.6  1998/08/20 19:48:27  basile
 * ok
 *
 * Revision 0.5  1998/08/18 19:33:34  basile
 * incomplete for important special keys
 *
 * Revision 0.4  1998/08/18 10:17:50  basile
 * shoudl have a toggled variable in generated code
 *
 * Revision 0.3  1998/08/17 18:48:22  basile
 * transition to more types
 *
 * Revision 0.2  1998/08/16 18:14:47  basile
 * parse a simplified fr-latin1 keymap
 *
 * Revision 0.1  1998/07/26 19:46:34  basile
 * loops in parsing
 * 
 *** endlog ****)

(* eof $Header: /home/basile/KUT/RCS/ut_kbdmap.ml,v 0.7 1998/08/22 19:48:10 basile Exp $ *)
